# Standard library imports
from enum import IntEnum
from array import array

# Third party imports

# Local application imports
from .bulk_message import BulkOutMessage


# Definitions
class PowerSelection(IntEnum):
    AutomaticVoltageSelection = 0x00
    Volts_5_0 = 0x01
    Volts_3_0 = 0x02
    Volts_1_8 = 0x03

    def __str__(self):
        match self:
            case PowerSelection.AutomaticVoltageSelection:
                return 'Automatic power selection'

            case PowerSelection.Volts_5_0:
                return '5V'

            case PowerSelection.Volts_3_0:
                return '3V'

            case PowerSelection.Volts_1_8:
                return '1.8V'


class PC_to_RDR_GetSlotStatus(BulkOutMessage):
    def __init__(self, bSlot: int, bSeq: int):
        msg = [0x65, 0x00, 0x00, 0x00, 0x00, bSlot, bSeq,
               0x00, 0x00, 0x00]
        super().__init__(array('B', msg))


class PC_to_RDR_IccPowerOn(BulkOutMessage):
    def __init__(self, bSlot: int, bSeq: int, bPowerSelect: PowerSelection):
        msg = [0x62, 0x00, 0x00, 0x00, 0x00, bSlot, bSeq,
               bPowerSelect, 0x00, 0x00]
        super().__init__(array('B', msg))

    @property
    def power_select(self):
        return PowerSelection(self.bPowerSelect)

    @property
    def bPowerSelect(self):
        return self.array[7]

    def __str__(self):
        return F"{super().__str__()}, {self.power_select.name}"


class PC_to_RDR_IccPowerOff(BulkOutMessage):
    def __init__(self, bSlot: int, bSeq: int):
        msg = [0x63, 0x00, 0x00, 0x00, 0x00, bSlot, bSeq, 0x00, 0x00, 0x00]
        super().__init__(array('B', msg))


class PC_to_RDR_Abort(BulkOutMessage):
    def __init__(self, bSlot: int, bSeq: int):
        msg = [0x72, 0x00, 0x00, 0x00, 0x00, bSlot, bSeq, 0x00, 0x00, 0x00]
        super().__init__(array('B', msg))


class PC_to_RDR_XfrBlock(BulkOutMessage):
    def __init__(self, bSlot: int, bSeq: int, bBWI: int, wLevelParameter: list[int], abData: list[int]):
        dwLength = list(int.to_bytes(len(abData), 4, byteorder='little'))
        msg = [0x6F] + dwLength + [bSlot, bSeq, bBWI] + \
            wLevelParameter + abData
        super().__init__(array('B', msg))


class PC_to_RDR_GetParameters(BulkOutMessage):
    def __init__(self, bSlot, bSeq):
        super().__init__(
            array('B', [0x6C, 0x00, 0x00, 0x00, 0x00, bSlot, bSeq, 0x00, 0x00, 0x00]))



"""binary.py
"""
# Standard library imports
from __future__ import annotations
from collections import UserString
import math
from typing import Callable, Iterable
from functools import reduce
from array import array
from operator import __add__, __xor__, __and__, __or__

# Third party imports

# Local application imports


# Helper functions
__HEX_DIGITS = '0123456789abcdefABCDEF'
__BIT_DIGITS = '01'
__WHITESPACE = '_'


def clean_char(c: str, keep: str, ignore: str, message: str) -> str:
    match c:
        case c if c in keep:
            return c
        case c if c in ignore:
            return ''
        case _:
            raise TypeError(
                F"Character '{c}' not allowed in {message}")


def clean_hex_char(c: str) -> str:
    return clean_char(c, keep=__HEX_DIGITS, ignore=__WHITESPACE, message="hexadecimal string")


def clean_bit_char(c: str) -> str:
    return clean_char(c, keep=__BIT_DIGITS, ignore=__WHITESPACE, message="binary string")


def bitwise_operation(left: str, right: str, operator: Callable[[int, int], int]):
    length = max(len(left), len(right))

    return format(operator(int(left, 16), int(right, 16)), F"0{length}X")


# 'HexString' class
class HexString(UserString):
    def __init__(self, string: str):
        value: str = reduce(__add__, map(clean_hex_char, string))
        super().__init__(value.upper())

    @property
    def bytes(self) -> bytes:
        return bytes.fromhex(self.data)

    @property
    def byte_length(self) -> int:
        if len(self) % 2 == 0:
            return len(self) // 2
        else:
            raise ValueError(F"{self.data} has an odd number of nibbles")

    @property
    def bit_string(self) -> BitString:
        return BitString(''.join([F"{int(nibble, 16):04b}" for nibble in self.data]))

    @property
    def bit_length(self) -> int:
        return len(self) * 4

    @property
    def int_list(self) -> list[int]:
        return list(self.bytes)

    @property
    def int_array(self) -> array:
        return array('B', self.bytes)

    @property
    def int(self) -> int:
        return int.from_bytes(self.bytes)

    @property
    def dscan_decimalize(self) -> str:
        """dscan_decimalize: double scan decimalization
        """
        dstr1 = ''
        dstr2 = ''
        for h in self.data:
            if h.isdigit():
                dstr1 = dstr1 + h
            else:
                dstr2 = dstr2 + F"{int(h, 16)-10}"

        return dstr1 + dstr2

    def blocks(self, bytesize: int) -> Iterable[HexString]:
        nr_blocks = math.ceil(self.byte_length / bytesize)
        return (self[i*2*bytesize:(i+1)*2*bytesize] for i in range(nr_blocks))

    def join(self, seq: Iterable[HexString]) -> HexString:
        return reduce(__add__, seq)

    def __or__(self, other: HexString) -> HexString:
        return HexString(bitwise_operation(self.data, other.data, __or__))

    def __xor__(self, other: HexString) -> HexString:
        return HexString(bitwise_operation(self.data, other.data, __xor__))

    def __and__(self, other: HexString) -> HexString:
        return HexString(bitwise_operation(self.data, other.data, __and__))

    def __invert__(self) -> HexString:
        return self ^ HexString('FF' * self.byte_length)


# 'ByteString' class
class ByteString(HexString):
    def __init__(self, string: str):
        if len(HexString(string)) % 2 != 0:
            super().__init__('0' + string)
        else:
            super().__init__(string)

    @property
    def byte_length(self) -> int:
        if len(self.data) % 2 == 0:
            return len(self.data) // 2
        else:
            raise ValueError(F"{self.data} has an odd number of nibbles")

    @property
    def bit_length(self) -> int:
        return len(self) * 8

    def blocks(self, bytesize: int) -> Iterable[ByteString]:
        nr_blocks = math.ceil(self.byte_length / bytesize)
        return (self[i*bytesize:(i+1)*bytesize] for i in range(nr_blocks))

    def join(self, seq: Iterable[ByteString]) -> ByteString:
        return reduce(__add__, seq)

    def __or__(self, other: ByteString) -> ByteString:
        return ByteString(bitwise_operation(self.data, other.data, __or__))

    def __xor__(self, other: ByteString) -> ByteString:
        return ByteString(bitwise_operation(self.data, other.data, __xor__))

    def __and__(self, other: ByteString) -> ByteString:
        return ByteString(bitwise_operation(self.data, other.data, __and__))

    def __invert__(self) -> ByteString:
        return self ^ ByteString('FF' * len(self))

    def __getitem__(self, key) -> ByteString:
        if isinstance(key, slice):
            return ByteString(self.bytes[key].hex())
        else:
            return ByteString(F"{self.bytes[key]:02X}")

    def __len__(self) -> int:
        return len(self.data) // 2


# 'BitString' class
class BitString(UserString):
    def __init__(self, string: str):
        value: str = ''.join(map(clean_bit_char, string))
        super().__init__(value.upper())

    @property
    def bytes(self) -> bytes:
        return HexString(self.data).bytes

    @property
    def byte_length(self) -> int:
        return HexString(self.data).byte_length

    @property
    def hex_string(self) -> HexString:
        if len(self) % 4 == 0:
            return HexString(format(int(self.data, 2), F"0{len(self)//4}X"))
        else:
            raise ValueError(
                F"{self.data} has a number of bit that is not a multiple of 4")

    @property
    def byte_string(self) -> ByteString:
        if len(self) % 8 == 0:
            return ByteString(format(int(self.data, 2), F"0{len(self)//4}X"))
        else:
            raise ValueError(
                F"{self.data} has a number of bit that is not a multiple of 4")

    @property
    def int_list(self) -> list[int]:
        return HexString(self.data).int_list

    def blocks(self, bitsize: int) -> Iterable[BitString]:
        nr_blocks = math.ceil(self.byte_length / bitsize)
        return (self[i*2*bitsize:(i+1)*2*bitsize] for i in range(nr_blocks))

    def join(self, seq: Iterable[BitString]) -> BitString:
        return reduce(__add__, seq)

    def permute(self, permutation: list[int]) -> BitString:
        str_out = BitString('')

        for i in permutation:
            str_out += self[i-1]

        return str_out

    def expand(self, expansion: list[int]) -> BitString:
        return self.permute(expansion)

    def left_circular_shit(self, shift: int) -> BitString:
        return self[shift:] + self[:shift]

    def __or__(self, other: BitString) -> BitString:
        return BitString(bitwise_operation(self.data, other.data, __or__))

    def __xor__(self, other: BitString) -> BitString:
        return BitString(bitwise_operation(self.data, other.data, __xor__))

    def __and__(self, other: BitString) -> BitString:
        return BitString(bitwise_operation(self.data, other.data, __and__))

    def __invert__(self) -> BitString:
        return self ^ BitString('1' * len(self))



"""emv_key_derivation.py
"""
# Standard library imports

# Third party imports
try:
    from icecream import ic
    ic.disable()
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa


# Local application imports
from common.binary import ByteString
from emv.key_management import master_key_derivation_A


# Example script
if __name__ == '__main__':
    ic.enable()
    ic.includeContext = True

    ic('Mastercard key derivation:')
    MASTERCARD_PAN = '5413123456784808'
    MASTERCARD_PSN = '00'
    MDK = ByteString('01234567899876543210012345678998')
    ic(MDK, MASTERCARD_PAN, MASTERCARD_PSN)
    MASTERCARD_UDK = master_key_derivation_A(
        MDK, MASTERCARD_PAN, MASTERCARD_PSN)
    ic(MASTERCARD_UDK)

    ic('Visa key derivation:')
    VISA_PAN = '4761739001010010'
    VISA_PSN = '00'

    ic(MDK, VISA_PAN, VISA_PSN)

    VISA_UDK = master_key_derivation_A(MDK, VISA_PAN, VISA_PSN)

    ic(VISA_UDK)



"""commands.py
"""

# Standard library imports
from enum import Enum, IntEnum
from typing import Union

# Third party imports

# Local application imports
from common.hstr import clean as _clean, to_intlist as _to_intlist
from globalplatform.encodings import SecureMessaging, CLA as GP_CLA
from iso7816 import CommandApdu, CommandField, CLA as ISO_CLA, Chaining, SecureMessaging as ISO_SecureMessaging


# Enum Definitions
class GetDataObject(Enum):
    ListOfApplications = '2F00'
    IssuerIdentificationNumber = '0042'
    CardImageNumber = '0045'
    CardData = '0066'
    KeyInformationTemplate = '00E0'
    CardCapabilityInformation = '0067'
    CurrentSecurityLevel = '00D3'
    SecurityDomainManagerURL = '5F50'
    ConfirmationCounter = '00C2'
    SequenceCounterOfTheDefaultKeyVersionNumber = '00C1'
    CardProductionLifeCycle = '9F7F'


class FileOccurrence(IntEnum):
    FirstOrOnlyOccurrence = 0x0
    NextOccurrence = 0x2


class ApplicationIdentifier(Enum):
    GlobalPlatformSecurityDomain = 'A000000151000000'
    Default = ''


# Command APDUs defined by GP
class GetData(CommandApdu):
    def __init__(self, secure_messaging: SecureMessaging, logical_channel: int, tag: Union[str, GetDataObject]):
        if isinstance(tag, GetDataObject):
            P1 = int(tag.value[0:2], 16)
            P2 = int(tag.value[2:4], 16)
        elif isinstance(tag, str):
            P1 = int(_clean(tag[0:2]), 16)
            P2 = int(_clean(tag[2:4]), 16)
        else:
            raise TypeError(
                F"globalplatform.GetData(): tag should be of type str or GetDataObject, received: {type(tag)}")

        super().__init__(GP_CLA(secure_messaging, logical_channel).value, 0xCA, P1, P2, Le=0x100)


def GET_DATA(secure_messaging: SecureMessaging, logical_channel: int, tag: Union[str, GetDataObject]) -> GetData:
    """GET_DATA: generate APDU for GET DATA command
    """
    return GetData(secure_messaging, logical_channel, tag)


class Select(CommandApdu):
    def __init__(self, logical_channel: int, file_occurrence: FileOccurrence, application_identifier: Union[str, ApplicationIdentifier]):
        cla = ISO_CLA(ISO_SecureMessaging.No,
                      Chaining.LastOrOnly, logical_channel).value
        ins = 0xA4
        p1 = 0x04
        p2 = file_occurrence.value
        Le = 0x100

        if isinstance(application_identifier, ApplicationIdentifier):
            if application_identifier == ApplicationIdentifier.Default:
                super().__init__(cla, ins, p1, p2, Le=Le)
            else:
                data = _to_intlist(application_identifier.value)
                super().__init__(cla, ins, p1, p2, data=data, Le=Le)

        elif isinstance(application_identifier, str):
            data = _to_intlist(application_identifier,
                               'SELECT()', 'application_identifier')
            super().__init__(cla, ins, p1, p2, data=data, Le=Le)

        else:
            raise TypeError(
                F"globalplatform.Select(): application_identifier should be of type str or ApplicationIdentifier, received: {type(application_identifier)}")


def SELECT(logical_channel: int, file_occurrence: FileOccurrence, application_identifier: Union[str, ApplicationIdentifier]) -> Select:
    return Select(logical_channel, file_occurrence, application_identifier)



"""encodings.py
"""

# Standard library imports
from enum import Enum

# Third party imports

# Local application imports
from common.hstr import clean as _clean


# Enum Definitions
class MultosDataField(Enum):
    MultosVersionNumber = "MULTOS Version Number"
    ICManufacturerIdentifier = "IC Manufacturer ID"
    ImplementerIdentifier = "Implementer ID"
    MCDIdentifier = "MCD ID"
    ProductIdentifier = "Product ID"
    IssuerIdentifier = "Issuer ID"
    MSMControlsDataDate = "MSM Controls Data Date"
    MCDNumber = "MCD Number"
    RFU = "RFU"
    MaximumDynamicSize = "Maximum Dynamic Size"
    MaximumPublicSize = "Maximum Public Size"
    MaximumDirectoryFileRecordSize = "Maximum DIR File Record Size"
    MaximumFCIRecordSize = "Maximum FCI Record Size"
    MaximumATRHistoricalByteRecordSize = "Maximum ATR Historical Byte Record Size"
    MaximumATRFileRecordSize = "Maximum ATR File Record Size"
    MultosPublicKeyCertificateLength = "MULTOS Public Key Certificate Length"
    SecurityLevel = "Security Level"
    CertificationMethodIdentifier = "Certification Method ID"
    ApplicationSignatureMethodIdentifier = "Application Signature Method ID"
    EnciphermentDescriptor = "Encipherment Descriptor"
    HashMethodIdentifier = "Hash Method ID"


# Class Definitions
class MultosData:
    def __init__(self, data: str):
        multos_data = _clean(
            data, "multos.encodings.MultosData.__init__()", "data")
        if len(multos_data) != 254:
            raise ValueError(
                F"MULTOS data should be 127 bytes, received {len(multos_data)//2} bytes: {multos_data}")
        else:
            self.__multos_data = multos_data

    def get_field(self, data: MultosDataField):
        pos, length = _MULTOS_DATA_DEFINITION[data]
        return self.__multos_data[2*pos:2*(pos+length)]

    def str(self):
        return self.__multos_data


# Helper functions
_MULTOS_DATA_DEFINITION = {
    MultosDataField.MultosVersionNumber: (0, 2),
    MultosDataField.ICManufacturerIdentifier: (2, 1),
    MultosDataField.ImplementerIdentifier: (3, 1),
    MultosDataField.MCDIdentifier: (4, 6),
    MultosDataField.ProductIdentifier: (10, 1),
    MultosDataField.IssuerIdentifier: (11, 4),
    MultosDataField.MSMControlsDataDate: (15, 1),
    MultosDataField.MCDNumber: (16, 8),
    MultosDataField.RFU: (24, 80),
    MultosDataField.MaximumDynamicSize: (104, 2),
    MultosDataField.MaximumPublicSize: (106, 2),
    MultosDataField.MaximumDirectoryFileRecordSize: (108, 2),
    MultosDataField.MaximumFCIRecordSize: (110, 2),
    MultosDataField.MaximumATRHistoricalByteRecordSize: (112, 2),
    MultosDataField.MaximumATRFileRecordSize: (114, 2),
    MultosDataField.MultosPublicKeyCertificateLength: (116, 2),
    MultosDataField.SecurityLevel: (118, 1),
    MultosDataField.CertificationMethodIdentifier: (119, 2),
    MultosDataField.ApplicationSignatureMethodIdentifier: (121, 2),
    MultosDataField.EnciphermentDescriptor: (123, 2),
    MultosDataField.HashMethodIdentifier: (125, 2)

}



"""commands.py
"""

# Standard library imports
from enum import StrEnum

# Third party imports

# Local application imports
# from common.ber import encode
from common.binary import ByteString
from iso7816.apdu import CommandApdu, CommandField, create_command_APDU


# Enum Definitions
class ApplicationIdentifier(StrEnum):
    Mastercard = 'A0000000041010'
    Visa = 'A0000000031010'
    Maestro = 'A00000000410103060'
    EVC = 'A0000008864040'
    EVC_test = 'A0000008863030'
    Dummy = 'A0000008862020'
    Dummy_test = 'A0000008861010'


class GetDataObject(StrEnum):
    ATC = '9F36'
    LastOnlineATCRegister = '9F13'
    PINTryCounter = '9F17'
    LogFormat = '9F4F'


# Command APDUs defined by EMV
class Select(CommandApdu):
    def __init__(self, aid: ByteString):
        super().__init__(ByteString(
            F'00A40400{len(aid):02X}') + aid + ByteString('00'))


def SELECT(aid: ByteString | ApplicationIdentifier) -> Select:
    """SELECT(): generate CommandApdu for SELECT command
    """
    if isinstance(aid, ApplicationIdentifier):
        return Select(ByteString(aid.value))
    elif isinstance(aid, ByteString):
        return Select(aid)
    else:
        raise TypeError(
            F"emv.SELECT(): aid should be of type ByteString or ApplicationIdentifier, received: {type(aid)}")


class GetProcessingOptions(CommandApdu):
    def __init__(self, pdol: ByteString):
        super().__init__(ByteString(
            F'80A80000{len(pdol):02X}') + pdol + ByteString('00'))


def GET_PROCESSING_OPTIONS(pdol: ByteString | None) -> GetProcessingOptions:
    """GET_PROCESSING_OPTIONS(): generate APDU for GET PROCESSING OPTIONS command
    """
    if pdol:
        return GetProcessingOptions(pdol)
    else:
        return GetProcessingOptions(ByteString('8300'))


GPO = GET_PROCESSING_OPTIONS


class ReadRecord(CommandApdu):
    def __init__(self, SFI: int, record: int):
        super().__init__(ByteString(F"00B2{record:02X}{SFI*8+4}00"))


def READ_RECORD(SFI: int, record: int) -> ReadRecord:
    """READ_RECORD(): generate APDU for READ RECORD command
    """
    return ReadRecord(SFI, record)


class GetData(CommandApdu):
    def __init__(self, tag: ByteString):
        super().__init__(ByteString(F'80CA{tag}00'))


def GET_DATA(tag: ByteString | GetDataObject) -> GetData:
    """GET_DATA: generate APDU for GET DATA command
    """
    if isinstance(tag, GetDataObject):
        if len(tag.value) == 2:
            P1 = ByteString('00')
            P2 = ByteString(tag.value)
        else:
            P1 = ByteString(tag.value[0:2])
            P2 = ByteString(tag.value[2:4])

    elif isinstance(tag, ByteString):
        P1 = tag[0:2]
        P2 = tag[2:4]

    else:
        raise TypeError(
            F"emv.GET_DATA(): tag should be of type ByteString or GetDataObject, received: {type(tag)}")

    return GetData(P1+P2)



# Standard library imports
from enum import IntEnum
from array import array
from copy import deepcopy
from typing import Union

# Third party imports

# Local application imports
from common.intlist import to_hstr as _to_hstr


# Definitions
class MessageType(IntEnum):
    PC_to_RDR_IccPowerOn = 0x62
    PC_to_RDR_IccPowerOff = 0x63
    PC_to_RDR_GetSlotStatus = 0x65
    PC_to_RDR_GetParameters = 0x6C
    PC_to_RDR_XfrBlock = 0x6F
    PC_to_RDR_Abort = 0x72
    RDR_to_PC_DataBlock = 0x80
    RDR_to_PC_SlotStatus = 0x81
    RDR_to_PC_Parameters = 0x82


class IccStatus(IntEnum):
    IccPresentAndActive = 0
    IccPresentAndInactive = 1
    NoIccPresent = 2
    RFU = 3


class CommandStatus(IntEnum):
    ProcessedWithoutError = 0
    Failed = 1
    TimeExtensionRequested = 2
    RFU = 3


class SlotError:
    def __init__(self, bError: int = None):
        self._bError = bError

    def __str__(self):
        error_str = {0xFF: "CMD_ABORTED (Host aborted the current activity)",
                     0xFE: "ICC_MUTE (CCID timed out while talking to the ICC)",
                     0xFD: "XFR_PARITY_ERROR (Parity error while talking to the ICC)",
                     0xFC: "XFR_OVERRUN (Overrun error while talking to the ICC)",
                     0xFB: "HW_ERROR (An all inclusive hardware error occurred)",
                     0xF8: "BAD_ATR_TS",
                     0xF7: "BAD_ATR_TCK",
                     0xF6: "ICC_PROTOCOL_NOT_SUPPORT ED",
                     0xF5: "ICC_CLASS_NOT_SUPPORTED",
                     0xF4: "PROCEDURE_BYTE_CONFLICT",
                     0xF3: "DEACTIVATED_PROTOCOL",
                     0xF2: "BUSY_WITH_AUTO_SEQUENCE (Automatic Sequence Ongoing)",
                     0xF0: "PIN_TIMEOUT",
                     0xEF: "PIN_CANCELLED",
                     0xE0: "CMD_SLOT_BUSY (A second command was sent to a slot which was already processing a command)",
                     0x00: "Command not supported"}.get(self._bError, None)

        if error_str is not None:
            return error_str
        elif self._bError >= 0x81 and self._bError <= 0xC0:
            return F"User defined error: {self._bError:02X}"
        elif self._bError >= 0x01 and self._bError <= 0x7F:
            return F"Unsupported or incorrect message parameter at index {self._bError:02X}"
        else:
            return "RFU"


class Protocol(IntEnum):
    T0 = 0x00
    T1 = 0x01
    TwoWire = 0x80
    ThreeWire = 0x81
    I2C = 0x82

    def __str__(self):
        match self:
            case Protocol.T0:
                return "T=0"
            case Protocol.T1:
                return "T=1"
            case _:
                return self.name


class BulkMessage:
    def __init__(self, message: Union[list[int], array]):
        if isinstance(message, array):
            self._message = deepcopy(message)
        elif isinstance(message, list):
            self._message = array('B', message)
        else:
            raise TypeError(
                F"Message should be array or list[int], received {type(message)}")

    @property
    def array(self) -> array:
        return deepcopy(self._message)

    @property
    def string(self) -> str:
        return _to_hstr(self._message)

    @property
    def header(self) -> str:
        return _to_hstr(self._message[0:10])

    @property
    def data(self) -> str:
        return _to_hstr(self._message[10:])

    @property
    def message_type(self) -> MessageType:
        return MessageType(self._message[0])

    @property
    def length(self) -> int:
        return int.from_bytes(self._message[1:5], byteorder='little')

    @property
    def bMessageType(self) -> int:
        return self._message[0]

    @property
    def dwLength(self) -> array:
        return self._message[1:5]

    @property
    def bSlot(self) -> int:
        return self._message[5]

    @property
    def bSeq(self) -> int:
        return self._message[6]

    def __str__(self):
        message_str = self.string
        content_hex = F"{message_str[0:2]} {message_str[2:10]} {message_str[10:12]} {message_str[12:14]} {message_str[14:]}"
        return F"{self.message_type.name} ({content_hex}): bSlot={message_str[10:12]}, bSeq={message_str[12:14]}"


class BulkOutMessage(BulkMessage):
    pass


class BulkInMessage(BulkMessage):
    @property
    def icc_status(self):
        return IccStatus(self.bStatus & 0x03)

    @property
    def command_status(self):
        return CommandStatus(self.bStatus >> 6)

    @property
    def error(self):
        if self.command_status == CommandStatus.Failed:
            return SlotError(self.bError)
        else:
            return SlotError()

    @property
    def bStatus(self):
        return self._message[7]

    @property
    def bError(self):
        return self._message[8]

    def __str__(self):
        if self.command_status == CommandStatus.Failed:
            return F"{super().__str__()}, {self.command_status.name}, {self.error}, {self.icc_status.name}"
        else:
            return F"{super().__str__()}, {self.command_status.name}, {self.icc_status.name}"



"""evc_dcvv.py
"""
# Standard library imports

# Third party imports
try:
    from icecream import ic
    ic.disable()
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa


# Local application imports
from common.binary import ByteString
from emv.key_management import master_key_derivation_A
from emv.dsc import generate_dcvv


# Example script
if __name__ == '__main__':
    ic.enable()
    ic.includeContext = True

    ic('Visa card data:')
    PAN = '4761739001010010'
    PSN = '00'
    EXPY_DATE = '1220'
    ic(PAN, PSN, EXPY_DATE)

    ic('Visa key derivation:')
    MDK = ByteString('2315208C9110AD402315208C9110AD40')
    ic(MDK)
    UDK = master_key_derivation_A(MDK, PAN, PSN)
    ic(UDK)

    ic('Visa dCVV computation:')
    for ATC in [F'{atc:04X}' for atc in range(1, 6)]:
        DCVV = generate_dcvv(UDK, PAN, ATC, EXPY_DATE)
        ic(ATC, DCVV)



"""modes.py: Block cipher modes of operation
"""

# Standard library imports
from typing import Callable, Iterable
from functools import reduce
from operator import __add__

# Third party imports

# Local application imports
from common.binary import ByteString


SymmetricCipher = Callable[[ByteString, ByteString], ByteString]


def ecb_encryption(enc: SymmetricCipher, key: ByteString, blocks: Iterable[ByteString]) -> ByteString:
    """ecb_encryption()
    """
    ciphertext_blocks = [enc(key, plaintext_block)
                         for plaintext_block in blocks]

    return reduce(__add__, ciphertext_blocks)


def ecb_decryption(dec: SymmetricCipher, key: ByteString, blocks: Iterable[ByteString]) -> ByteString:
    """ecb_decryption()
    """
    plaintext_blocks = [dec(key, ciphertext_block)
                        for ciphertext_block in blocks]

    return reduce(__add__, plaintext_blocks)


def cbc_encryption(enc: SymmetricCipher, key: ByteString, blocks: Iterable[ByteString], iv: ByteString) -> ByteString:
    """cbc_encryption()
    """
    tmp = iv
    ciphertext_blocks: list[ByteString] = []
    for plaintext_block in blocks:
        tmp = enc(key, tmp ^ plaintext_block)
        ciphertext_blocks.append(tmp)

    return reduce(__add__, ciphertext_blocks)


def cbc_decryption(dec: SymmetricCipher, key: ByteString, blocks: Iterable[ByteString], iv: ByteString) -> ByteString:
    """cbc_decryption()
    """
    tmp = iv
    plaintext_blocks: list[ByteString] = []
    for ciphertext_block in blocks:
        plaintext_blocks.append(tmp ^ dec(key, ciphertext_block))
        tmp = ciphertext_block

    return reduce(__add__, plaintext_blocks)



"""apdu.py
"""

# Standard library imports
from enum import Enum
from copy import deepcopy
from array import array

# Third party imports

# Local application imports
from common.hstr import clean as _clean, to_intlist as _to_intlist
from common.intlist import to_hstr as _to_hstr, is_int_list as _is_int_list


# Enum Definitions
class CommandCase(Enum):
    Case1 = 'Case 1'
    Case2S = 'Case 2S'
    Case2E = 'Case 2E'
    Case3S = 'Case 3S'
    Case3E = 'Case 3E'
    Case4S = 'Case 4S'
    Case4E = 'Case 4E'


class CommandField(Enum):
    Header = 'Header'
    Body = 'Body'
    Class = 'Class'
    Instruction = 'Instruction'
    P1 = 'P1'
    P2 = 'P2'
    Lc = 'Lc'
    Data = 'Data'
    Le = 'Le'


class ResponseField(Enum):
    Body = 'Body'
    Trailer = 'Trailer'
    Data = 'Data'
    SW1 = 'SW1'
    SW2 = 'SW2'
    SW12 = 'SW12'


class ResponseProcessingState(Enum):
    Normal = 'Normal procesing'
    Warning = 'Warning processing'
    ExecutionError = 'Execution error'
    CheckingError = 'Checking error'


# Command and Response APDU classes
class CommandApdu:
    # type: ignore
    def __init__(self, CLA: int, INS: int, P1: int, P2: int, data: list[int] = None, Le: int = None):
        if CLA < 0 or CLA > 255:
            raise ValueError(
                F"CommandApdu(): CLA should be [0-255], received: {CLA}")
        elif INS < 0 or INS > 255:
            raise ValueError(
                F"CommandApdu(): INS should be [0-255], received: {INS}")
        elif P1 < 0 or P1 > 255:
            raise ValueError(
                F"CommandApdu(): P1 should be [0-255], received: {P1}")
        elif P2 < 0 or P2 > 255:
            raise ValueError(
                F"CommandApdu(): P2 should be [0-255], received: {P2}")
        elif data is not None:
            if not _is_int_list(data):
                raise ValueError(
                    F"CommandApdu(): data should be list of [0-255], received: {data}")

        self.__content = array('B', [CLA, INS, P1, P2])

        match data, Le:
            case None, None:
                # Case 1
                self.__case = CommandCase.Case1

            case None, _:
                # Case 2
                if Le <= 0:
                    raise ValueError(
                        F"CommandApdu(): Le should be [0-65,536], received: {Le}")
                elif Le < 256:
                    # Case 2S
                    self.__case = CommandCase.Case2S
                    self.__content.append(Le)
                elif Le == 256:
                    # Case 2S
                    self.__case = CommandCase.Case2S
                    self.__content.append(0x00)
                elif Le < 65536:
                    # Case 2E
                    self.__case = CommandCase.Case2E
                    self.__content.extend(int.to_bytes(Le, 3, byteorder='big'))
                elif Le == 65536:
                    # Case 2E
                    self.__case = CommandCase.Case2E
                    self.__content.extend([0x00, 0x00, 0x00])
                else:
                    raise ValueError(
                        F"CommandApdu(): Le should be [0-65,536], received: {Le}")

            case _, None:
                # Case 3
                Lc = len(data)
                if Lc == 0:
                    raise ValueError(
                        F"CommandApdu(): data should not be empty, received: {data}")
                elif Lc < 256:
                    # Case 3S
                    self.__case = CommandCase.Case3S
                    self.__content.append(Lc)
                    self.__content.extend(data)
                elif Lc < 65536:
                    # Case 3E
                    self.__case = CommandCase.Case3E
                    self.__content.extend(int.to_bytes(Lc, 3, byteorder='big'))
                else:
                    raise ValueError(
                        F"CommandApdu(): length of data should be [1-65,535] bytes, received {Lc} bytes ({data})")

            case _, _:
                # Case 4
                Lc = len(data)
                if Le <= 0:
                    raise ValueError(
                        F"CommandApdu(): Le should be [0-65,536], received: {Le}")
                elif Lc == 0:
                    raise ValueError(
                        F"CommandApdu(): data should not be empty, received: {data}")
                elif Lc < 256 and Le <= 256:
                    # Case 4S
                    self.__case = CommandCase.Case4S
                    self.__content.append(Lc)
                    self.__content.extend(data)
                    if Le == 256:
                        self.__content.append(0x00)
                    else:
                        self.__content.append(Le)
                elif Lc < 65536 and Le <= 65536:
                    # Case 4E
                    self.__case = CommandCase.Case4E
                    self.__content.extend(int.to_bytes(Lc, 3, byteorder='big'))
                    self.__content.extend(data)
                    if Le == 65536:
                        self.__content.extend([0x00, 0x00])
                    else:
                        self.__content.extend(
                            int.to_bytes(Le, 2, byteorder='big'))

    @classmethod
    def from_string(cls, apdu_str: str):
        _apdu_list = _to_intlist(
            apdu_str, "CommandApdu.__init__()", "apdu_str")

        apdu_length = len(_apdu_list)

        if apdu_length < 4:
            raise ValueError(
                F'Wrong Command APDU length: expected 4 or more bytes but received {apdu_length}')

        elif apdu_length == 4:
            # Case 1
            return cls(_apdu_list[0], _apdu_list[1], _apdu_list[2], _apdu_list[3])

        elif apdu_length == 5:
            # Case 2S
            return cls(_apdu_list[0], _apdu_list[1], _apdu_list[2], _apdu_list[3], Le=_apdu_list[4])

        elif _apdu_list[4] != 0x00:
            if apdu_length == 5 + _apdu_list[4]:
                # Case 3S
                return cls(_apdu_list[0], _apdu_list[1], _apdu_list[2], _apdu_list[3], data=_apdu_list[5:])
            elif apdu_length == 5 + _apdu_list[4] + 1:
                # Case 4S
                return cls(_apdu_list[0], _apdu_list[1], _apdu_list[2], _apdu_list[3], data=_apdu_list[5:-1], Le=_apdu_list[-1])
            else:
                raise ValueError(
                    F'Wrong Command APDU length: expected {5+_apdu_list[4]} bytes for Case 3s or {5+_apdu_list[4]+1} bytes for Case 4s but received {apdu_length} bytes instead')

        else:
            # Extended length cases
            if apdu_length == 7:
                # Case 2E
                if int.from_bytes(_apdu_list[4:], byteorder='big') == 0:
                    return cls(_apdu_list[0], _apdu_list[1], _apdu_list[2], _apdu_list[3], Le=65536)
                else:
                    return cls(_apdu_list[0], _apdu_list[1], _apdu_list[2], _apdu_list[3], Le=int.from_bytes(_apdu_list[4:], byteorder='big'))

            elif apdu_length == 7 + _apdu_list[5] * 0x0100 + _apdu_list[6]:
                # Case 3E
                return cls(_apdu_list[0], _apdu_list[1], _apdu_list[2], _apdu_list[3], data=_apdu_list[7:])

            elif apdu_length == 7 + _apdu_list[5] * 0x0100 + _apdu_list[6] + 2:
                # Case 4E
                return cls(_apdu_list[0], _apdu_list[1], _apdu_list[2], _apdu_list[3], data=_apdu_list[5:-2], Le=int.from_bytes(_apdu_list[-2:], byteorder='big'))

            else:
                raise ValueError(
                    F'Wrong Command APDU length: expected {7 + _apdu_list[5] * 0x0100 + _apdu_list[6]} bytes for Case 3e or {7 + _apdu_list[5] * 0x0100 + _apdu_list[6] + 2} bytes for Case 4e but received {apdu_length} bytes instead')

    @property
    def case(self) -> CommandCase:
        return self.__case

    @property
    def list(self) -> list[int]:
        return list(self.__content)

    @property
    def string(self) -> str:
        return _to_hstr(self.__content)

    @property
    def array(self):
        return deepcopy(self.__content)

    @property
    def header(self):
        return self.__content[0:4]

    @property
    def body(self):
        return self.__content[4:]

    @property
    def CLA(self):
        return self.__content[0]

    @property
    def INS(self):
        return self.__content[1]

    @property
    def P1(self):
        return self.__content[2]

    @property
    def P2(self):
        return self.__content[3]

    @property
    def Lc(self):
        match self.__case:
            case CommandCase.Case1 | CommandCase.Case2S | CommandCase.Case2E:
                raise ValueError(
                    F'Command APDU has no Lc field ({self.case.value})')
            case CommandCase.Case3S | CommandCase.Case4S | CommandCase.Case3E | CommandCase.Case4E:
                return len(self.data)

    @property
    def data(self):
        match self.case:
            case CommandCase.Case1 | CommandCase.Case2S | CommandCase.Case2E:
                raise ValueError(
                    F'Command APDU has no Data field ({self.case.value})')
            case CommandCase.Case3S:
                return self.__content[5:]
            case CommandCase.Case4S:
                return self.__content[5:-1]
            case CommandCase.Case3E:
                return self.__content[7:]
            case CommandCase.Case4E:
                return self.__content[7:-2]

    @property
    def Le(self):
        match self.case:
            case CommandCase.Case1 | CommandCase.Case3S | CommandCase.Case3E:
                raise ValueError(
                    F'Command APDU has no Le field ({self.case.value})')
            case CommandCase.Case2S | CommandCase.Case4S:
                if self.__content[-1] == 0x00:
                    return 256
                else:
                    return self.__content[-1]
            case CommandCase.Case2E | CommandCase.Case4E:
                if list(self.__content[-2:]) == [0x00, 0x00]:
                    return 65536
                else:
                    return int.from_bytes(self.__content[-2:], byteorder='big')

    def __str__(self):
        str = F"{self.__class__.__name__} ({self.case.value}): CLA={self.CLA:02X}, INS={self.INS:02X}, P1={self.P1:02X}, P2={self.P2:02X}"

        match self.case:
            case CommandCase.Case1:
                return str
            case CommandCase.Case2S:
                return str + F", Le={self.Le:02X}"
            case CommandCase.Case2E:
                return str + F", Le={self.Le:06X}"
            case CommandCase.Case3S:
                return str + F", Lc={self.Lc:02X}, DATA={_to_hstr(self.data)}"
            case CommandCase.Case3E:
                return str + F", Lc={self.Lc:06X}, DATA={_to_hstr(self.data)}"
            case CommandCase.Case4S:
                return str + F", Lc={self.Lc:02X}, DATA={_to_hstr(self.data)}, Le={self.Le:02X}"
            case CommandCase.Case4E:
                return str + F", Lc={self.Lc:06X}, DATA={_to_hstr(self.data)}, Le={self.Le:04X}"

    def update_Le(self, Le: str):
        match self.case:
            case CommandCase.Case1 | CommandCase.Case3S | CommandCase.Case3E:
                raise ValueError(
                    F'Command APDU has no Le field ({self.case.value})')

            case CommandCase.Case2S | CommandCase.Case2E:
                return CommandApdu(self.CLA, self.INS, self.P1, self.P2, Le=int(Le, 16))

            case CommandCase.Case4S | CommandCase.Case4E:
                return CommandApdu(self.CLA, self.INS, self.P1, self.P2, data=list(self.data), Le=int(Le, 16))

    def with_updated_Le(self, Le: int):
        match self.case:
            case CommandCase.Case1 | CommandCase.Case3S | CommandCase.Case3E:
                raise ValueError(
                    F'Command APDU has no Le field ({self.case.value})')

            case CommandCase.Case2S | CommandCase.Case2E:
                return CommandApdu(self.CLA, self.INS, self.P1, self.P2, Le=Le)

            case  CommandCase.Case4S | CommandCase.Case4E:
                return CommandApdu(self.CLA, self.INS, self.P1, self.P2, data=list(self.data), Le=Le)


class StatusBytes:
    def __init__(self, sw12: str):
        self.__list = _to_intlist(sw12, "StatusBytes.__init__()", "sw12")
        if len(self.__list) != 2:
            raise ValueError(F"Expected 2 bytes, received: {sw12}")

        first_digit = self.__list[0] >> 4
        if first_digit not in [0x6, 0x9]:
            raise ValueError(F"Expected 6XXX or 9XXX value, received: {sw12}")

        if self.__list[0] == 0x60:
            raise ValueError(F"60XX in valid, received: {sw12}")

        match self.__list:
            case [0x90, 0x00]:
                self.__state = ResponseProcessingState.Normal
                self.__meaning = 'No further qualification'
            case [0x61, _]:
                self.__state = ResponseProcessingState.Normal
                self.__meaning = 'SW2 encodes the number of data bytes still available'
            case [0x62, sw2]:
                self.__state = ResponseProcessingState.Warning
                self.__meaning = {0x00: 'No information given',
                                  0x81: 'Part of returned data may be corrupted',
                                  0x82: 'End of file or record reached before reading Ne bytes',
                                  0x83: 'Selected file deactivated',
                                  0x84: 'File control information not formatted according to ISO7816-4 5.3.3',
                                  0x85: 'Selected file in termination state',
                                  0x86: 'No input data available from a sensor on the card'}.get(sw2, 'State of non-volatile memory is unchanged')
            case [0x63, sw2]:
                self.__state = ResponseProcessingState.Warning
                self.__meaning = {0x00: 'No information given',
                                  0x81: 'File filled up by the last '}.get(sw2, 'State of non-volatile memory has changed(further qualification in SW2)')
            case [0x64, sw2]:
                self.__state = ResponseProcessingState.ExecutionError
                self.__meaning = {0x00: 'Execution error',
                                  0x01: 'Immediate response required by the card'}.get(sw2, 'State of non-volatile memory is unchanged (further qualification in SW2)')
            case [0x65, sw2]:
                self.__state = ResponseProcessingState.ExecutionError
                self.__meaning = {0x00: 'No information given',
                                  0x81: 'Memory failure'}.get(sw2, 'State of non-volatile memory has changed (further qualification in SW2)')
            case [0x66, _]:
                self.__state = ResponseProcessingState.ExecutionError
                self.__meaning = 'Security-related issues'
            case [0x67, 0x00]:
                self.__state = ResponseProcessingState.CheckingError
                self.__meaning = 'Wrong length; no further indication'
            case [0x68, sw2]:
                self.__state = ResponseProcessingState.CheckingError
                self.__meaning = {0x00: 'No information given',
                                  0x81: 'Logical channel not supported',
                                  0x82: 'Secure messaging not supported',
                                  0x83: 'Last command of the chain expected',
                                  0x84: 'Command chaining not supported'}.get(sw2, 'Functions in CLA not supported (further qualification in SW2)')
            case [0x69, sw2]:
                self.__state = ResponseProcessingState.CheckingError
                self.__meaning = {0x00: 'No information given',
                                  0x81: 'Command incompatible with file structure',
                                  0x82: 'Security status not satisfied',
                                  0x83: 'Authentication method blocked',
                                  0x84: 'Reference data not usable',
                                  0x85: 'Conditions of use not satisfied',
                                  0x86: 'Command not allowed (no current EF)',
                                  0x87: 'Expected secure messaging data objects missing',
                                  0x88: 'Incorrect secure messaging data objects'}.get(sw2, 'Command not allowed (further qualification in SW2)')
            case [0x6A, _]:
                self.__state = ResponseProcessingState.CheckingError
                self.__meaning = 'Wrong parameters P1-P2 (further qualification in SW2)'
            case [0x6B, 0x00]:
                self.__state = ResponseProcessingState.CheckingError
                self.__meaning = 'Wrong parameters P1-P2'
            case [0x6C, _]:
                self.__state = ResponseProcessingState.CheckingError
                self.__meaning = 'Wrong Le field; SW2 encodes the exact number of available data bytes'
            case [0x6D, 0x00]:
                self.__state = ResponseProcessingState.CheckingError
                self.__meaning = 'Instruction code not supported or invalid'
            case [0x6E, 0x00]:
                self.__state = ResponseProcessingState.CheckingError
                self.__meaning = 'Class not supported'
            case [0x6F, 0x00]:
                self.__state = ResponseProcessingState.CheckingError
                self.__meaning = 'No precise diagnosis'

    @property
    def state(self):
        return self.__state

    @property
    def meaning(self):
        return self.__meaning


class ResponseApdu:
    def __init__(self, apdu_str: str):
        self.__list = _to_intlist(
            apdu_str, "ResponseApdu.__init__()", "apdu_str")

    @classmethod
    def from_list(cls, response: list[int]):
        return cls(_to_hstr(response))

    @property
    def list(self):
        return self.__list

    @property
    def string(self):
        return _to_hstr(self.__list)

    def get_field(self, field: ResponseField) -> str:
        match field:
            case ResponseField.Trailer | ResponseField.SW12:
                return _to_hstr(self.__list[-2:])

            case ResponseField.SW1:
                return F"{self.__list[-2]:02X}"

            case ResponseField.SW2:
                return F"{self.__list[-1]:02X}"

            case ResponseField.Body | ResponseField.Data:
                if len(self.__list) > 2:
                    return _to_hstr(self.__list[0:-2])
                else:
                    raise ValueError(
                        'Wrong Response APDU field: no response body')

    @property
    def data(self) -> str:
        return self.get_field(ResponseField.Data)

    @property
    def SW12(self) -> str:
        return self.get_field(ResponseField.SW12)

    @property
    def SW1(self) -> str:
        return self.get_field(ResponseField.SW1)

    @property
    def SW2(self) -> str:
        return self.get_field(ResponseField.SW2)

    @property
    def StatusBytes(self) -> StatusBytes:
        return StatusBytes(self.get_field(ResponseField.SW12))

    def __str__(self):
        return _to_hstr(self.__list)


#
# Helper functions
#
def _Lc(data: list[int]):
    Lc = len(data)
    if Lc < 256:
        return [Lc]

    if Lc < 65536:
        return [0x00, Lc >> 8, Lc & 0x00FF]

    raise ValueError(F'Data length should be <65,536: actual length = {Lc}')


def _dict_to_string(apdu: dict):
    # If 'apdu' includes a header, it will be used directly. If not, the header will be constructed with CLA, INS, P1, and P2
    if CommandField.Header in apdu:
        header = apdu[CommandField.Header]
    else:
        header = apdu[CommandField.Class] + apdu[CommandField.Instruction] + \
            apdu[CommandField.P1] + apdu[CommandField.P2]

    # If 'apdu' includes a body, it will be used directly. If not, the body will be constructed with the optional DATA and Le fields
    body = ''
    if CommandField.Body in apdu:
        body = apdu[CommandField.Body]
    else:
        if CommandField.Data in apdu:
            _DATA = apdu[CommandField.Data]
            body = F"{''.join([F'{l:02X}' for l in _Lc(_to_intlist(_DATA))])}{_DATA}"

        if CommandField.Le in apdu:
            body += apdu[CommandField.Le]

    return header + body


def _byte_to_string(byte, command_name='()', byte_name='') -> str:
    if (byte < 0x00) or (byte > 0xFF):
        raise ValueError(
            F'{command_name}: {byte_name} out of bound, received {byte:X}')
    else:
        return F"{byte:02X}"



"""encodings.py
"""

# Standard library imports
from enum import Enum

# Third party imports

# Local application imports
from common.hstr import clean as _clean


# Enum Definitions
class SecureMessaging(Enum):
    No = 'No SM or no indication'
    Proprietary = 'Proprietary SM format'
    Iso7816 = 'SM according to ISO7816-6 6, command header not processed'
    Iso7816_CMAC = 'SM according to ISO7816-6 6, command header authenticated'


class Chaining(Enum):
    LastOrOnly = 'The command is the last or only command of a chain'
    NotLast = 'The command is not the last command of a chain'


# Class Definitions
class CLA:
    def __init__(self, secure_messaging: SecureMessaging, chaining: Chaining, logical_channel: int):
        if (logical_channel < 0) or (logical_channel > 19):
            raise ValueError(
                F'Logical channel number out of bound: should be in [0;19], received {logical_channel}')

        self.__logical_channel: int = logical_channel
        self.__chaining: Chaining = chaining
        self.__secure_messaging: SecureMessaging = secure_messaging

    @classmethod
    def from_value(cls, CLA: int):
        if (CLA < 0x00) or (CLA > 0xFF):
            raise ValueError(
                F'Class byte out of bound: should be 2 bytes, received {CLA:X}')

        secure_messaging: SecureMessaging

        chaining: Chaining
        if CLA & 0x10 == 0x00:
            chaining = Chaining.LastOrOnly
        else:
            chaining = Chaining.NotLast

        logical_channel: int

        # Testing b7
        if CLA & 0x40 == 0x00:
            # CLA Byte Coding according to Table 2
            # Testing b4-b3
            match CLA & 0x0C:
                case 0x00:
                    secure_messaging = SecureMessaging.No
                case 0x04:
                    secure_messaging = SecureMessaging.Proprietary
                case 0x08:
                    secure_messaging = SecureMessaging.Iso7816
                case 0x0C:
                    secure_messaging = SecureMessaging.Iso7816_CMAC
                case _:
                    # This should not happen
                    raise ArithmeticError(
                        F'CLA.from_value() error: CLA = {CLA:02X}')

            logical_channel = CLA & 0x03

        else:
            # CLA Byte Coding according to Table 3
            # Testing b6
            if CLA & 0x20 == 0x00:
                secure_messaging = SecureMessaging.No
            else:
                secure_messaging = SecureMessaging.Iso7816

            logical_channel = CLA & 0x0F

        return cls(secure_messaging, chaining, logical_channel)

    @classmethod
    def from_string(cls, CLA: str):
        cla = _clean(CLA, 'from_string()', 'CLA')
        return cls.from_value(int(cla, base=16))

    @property
    def value(self) -> int:
        CLA = 0x00

        if self.__logical_channel <= 3:
            # CLA Byte Coding according to Table 11-11
            CLA += self.__logical_channel
            match self.__secure_messaging:
                case SecureMessaging.No:
                    CLA += 0x00
                case SecureMessaging.Proprietary:
                    CLA += 0x04
                case SecureMessaging.Iso7816:
                    CLA += 0x08
                case SecureMessaging.Iso7816_CMAC:
                    CLA += 0x0C

        else:
            # CLA Byte Coding according to Table 11-12
            CLA += 0x40 + self.__logical_channel
            if self.__secure_messaging != SecureMessaging.No:
                CLA += 0x20

        if self.__chaining == Chaining.NotLast:
            CLA += 0x10

        return CLA

    @property
    def string(self) -> str:
        return F"{self.value:02X}"



"""commands.py
"""

# Standard library imports
from enum import IntEnum

# Third party imports

# Local application imports
from common.hstr import clean as _clean
from iso7816.apdu import CommandApdu
from iso7816.encodings import SecureMessaging, Chaining, CLA


# Enum Definitions
class Selection(IntEnum):
    SelectMForDForEF = 0x00
    SelectChildDF = 0x01
    SelectEFUnderCurrentDF = 0x02
    SelectParentDFOfCurrentDF = 0x03
    SelectByDFName = 0x04
    SelectFromMF = 0x08
    SelectFromCurrentDF = 0x09


class FileOccurrence(IntEnum):
    FirstOrOnlyOccurrence = 0x0
    LastOccurrence = 0x1
    NextOccurrence = 0x2
    PreviousOccurrence = 0x3


class FileControlInformation(IntEnum):
    ReturnFCITemplate = 0x00
    ReturnFCPTempalte = 0x10
    ReturnFMDTemplate = 0x20
    NoResponseOrProprietary = 0x30


#
# Commands for interchange
#
class GetResponse(CommandApdu):
    def __init__(self, CLA: CLA, Le: int):
        super().__init__(CLA.value, 0xC0, 0x00, 0x00, Le=Le)


def GET_RESPONSE(CLA: CLA, Le: int) -> GetResponse:
    return GetResponse(CLA, Le)


class Select(CommandApdu):
    def __init__(self,
                 secure_messaging: SecureMessaging,
                 chaining: Chaining,
                 logical_channel: int,
                 selection: Selection,
                 file_occurrence: FileOccurrence,
                 fci: FileControlInformation,
                 data: list[int] = None,
                 Le: int = None):
        cla: int = CLA(secure_messaging, chaining, logical_channel).value
        INS: int = 0xA4
        P1: int = selection.value
        P2: int = file_occurrence.value + fci.value

        super().__init__(cla, INS, P1, P2, data=data, Le=Le)


def SELECT(secure_messaging: SecureMessaging,
           chaining: Chaining,
           logical_channel: int,
           selection: Selection,
           file_occurrence: FileOccurrence,
           fci: FileControlInformation,
           data: list[int] = None,
           Le: int = None) -> Select:
    return Select(secure_messaging, chaining, logical_channel, selection, file_occurrence, fci, data, Le)



"""test_crypto_des.py
"""

# Standard library imports
import unittest

# Local application imports
from crypto.des import dea_e, dea_d, dea_ede_cbc, tdea_2_ede, tdea_2_ded, tdea_2_ede_ecb, tdea_2_ded_ecb, tdea_2_ede_cbc, tdea_2_ded_cbc, adjust_parity
from common.binary import HexString


#
# Test values
#
key4_0_to_7 = HexString('01234567')

key8_0_to_F = HexString('0123456789ABCDEF')
key8_0_to_F_1 = HexString('0123_4567_89AB_CDEF')
key8_0_to_F_5 = HexString('01234567_89ABCDEF')

key16_0_to_F_to_0 = HexString('0123456789ABCDEFFEDCBA9876543210')
key16_0_to_F_to_0_1 = HexString('0123456789ABCDEF_FEDCBA9876543210')

zeroes7 = HexString('00' * 7)
zeroes8 = HexString('00' * 8)
zeroes15 = HexString('00' * 15)
zeroes16 = HexString('00' * 16)

CC8 = HexString('CC' * 8)

FF8 = HexString('FF' * 8)
FF16 = HexString('FF' * 16)

fortyfour16 = HexString('44' * 16)


#
# Unit tests
#
class TestMethods(unittest.TestCase):
    """Unit tests for 'des' module
    """

    def test_dea_e(self):
        self.assertEqual(dea_e(key8_0_to_F, zeroes8),
                         'D5D44FF720683D0D')
        self.assertEqual(dea_e(key8_0_to_F_1, zeroes8),
                         'D5D44FF720683D0D')
        with self.assertRaises(ValueError):
            dea_e(key8_0_to_F, zeroes7)
        with self.assertRaises(ValueError):
            dea_e(key4_0_to_7, zeroes8)

    def test_dea_d(self):
        self.assertEqual(dea_d(key8_0_to_F, zeroes8),
                         '14AAD7F4DBB4E094')
        self.assertEqual(dea_d(key8_0_to_F_1, zeroes8),
                         '14AAD7F4DBB4E094')

    def test_dea_ede_cbc(self):
        self.assertEqual(dea_ede_cbc(key8_0_to_F, fortyfour16, zeroes8),
                         '8850517E7A817B1EBA5E2B72B6217B50')
        self.assertEqual(dea_ede_cbc(key8_0_to_F_5, fortyfour16, CC8),
                         'F31C939892FEFC8F14DBA3B2C7BAF0A7')
        with self.assertRaises(ValueError):
            dea_ede_cbc(key8_0_to_F, zeroes15, zeroes8)
        with self.assertRaises(ValueError):
            dea_ede_cbc(key8_0_to_F, zeroes16, zeroes7)

    def test_tdea_2_ede(self):
        self.assertEqual(tdea_2_ede(key16_0_to_F_to_0, zeroes8),
                         '08D7B4FB629D0885')
        self.assertEqual(tdea_2_ede(key16_0_to_F_to_0_1, zeroes8),
                         '08D7B4FB629D0885')

    def test_tdea_2_ded(self):
        self.assertEqual(tdea_2_ded(key16_0_to_F_to_0, zeroes8),
                         'C1E6E95D2166B5C4')
        self.assertEqual(tdea_2_ded(key16_0_to_F_to_0_1, zeroes8),
                         'C1E6E95D2166B5C4')

    def test_tdea_2_ede_ecb(self):
        self.assertEqual(tdea_2_ede_ecb(HexString('F19DADD9AEA429D92FB5F7B92A15FD04'),
                         HexString('01234567899876543210012345678998')),
                         '7E3F0B0E968FA631C8E4618CA317256A')
        self.assertEqual(tdea_2_ede_ecb(HexString('F19DADD9AEA429D92FB5F7B92A15FD04'),
                         HexString('01017391010101101320010101010101')),
                         '0AA72BC2B91EFDCFFFD0EDD21A03F200')

    def test_tdea_2_ded_ecb(self):
        self.assertEqual(tdea_2_ded_ecb(HexString('F19DADD9AEA429D92FB5F7B92A15FD04'),
                         HexString('7E3F0B0E968FA631C8E4618CA317256A')),
                         '01234567899876543210012345678998')
        self.assertEqual(tdea_2_ded_ecb(HexString('F19DADD9AEA429D92FB5F7B92A15FD04'),
                         HexString('0AA72BC2B91EFDCFFFD0EDD21A03F200')),
                         '01017391010101101320010101010101')

    def test_tdea_2_ede_cbc(self):
        self.assertEqual(tdea_2_ede_cbc(key16_0_to_F_to_0, zeroes16, zeroes8),
                         '08D7B4FB629D08850A121DC33DFB5947')
        self.assertEqual(tdea_2_ede_cbc(key16_0_to_F_to_0_1, zeroes16, FF8),
                         '847CA792BFA6FD4C9154099C397A0ABA')
        with self.assertRaises(ValueError):
            tdea_2_ede_cbc(key16_0_to_F_to_0, zeroes15, zeroes8)
        with self.assertRaises(ValueError):
            tdea_2_ede_cbc(key16_0_to_F_to_0, zeroes16, zeroes7)

    def test_tdea_2_ded_cbc(self):
        self.assertEqual(tdea_2_ded_cbc(key16_0_to_F_to_0, zeroes16, zeroes8),
                         'C1E6E95D2166B5C4C1E6E95D2166B5C4')
        self.assertEqual(tdea_2_ded_cbc(key16_0_to_F_to_0_1, zeroes16, FF8),
                         '3E1916A2DE994A3BC1E6E95D2166B5C4')
        with self.assertRaises(ValueError):
            tdea_2_ded_cbc(key16_0_to_F_to_0, zeroes15, zeroes8)
        with self.assertRaises(ValueError):
            tdea_2_ded_cbc(key16_0_to_F_to_0, zeroes16, zeroes7)

    def test_adjust_parity(self):
        self.assertEqual(adjust_parity(key16_0_to_F_to_0), key16_0_to_F_to_0)
        self.assertEqual(adjust_parity(HexString('462EC416E0E83C04_2CD1B10731AB4736')),
                         '462FC416E0E93D042CD0B00731AB4637')


if __name__ == '__main__':
    unittest.main()



"""dsc.py: Dynamic Security Codes module
"""

# Standard library imports

# Third party imports
try:
    from icecream import ic
    ic.disable()
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa

# Local application imports
from common import hstr
from common.binary import ByteString
from crypto import des


# Function definitions
def generate_dcvv(udk: ByteString, pan: str, atc: str, expiration_date: str):
    alternate_pan = F"{atc}{pan[4:]}"
    ic(alternate_pan)

    input_block = F"{alternate_pan}{expiration_date}"
    input_block = ByteString(input_block + '0' * (32 - len(input_block)))

    udk_a = udk[0:8]
    udk_b = udk[8:16]
    block_a = input_block[0:8]
    block_b = input_block[8:16]
    ic(block_a, block_b)

    block_c = des.dea_e(udk_a, block_a)
    block_d = block_c | block_b
    block_e = des.dea_e(udk_a, block_d)
    block_f = des.dea_d(udk_b, block_e)
    block_g = des.dea_e(udk_a, block_f)
    block_h = block_g.dscan_decimalize
    ic(block_c, block_d, block_e, block_f, block_g, block_h)

    return block_h[0:3]


def generate_ivcvc3(udk: ByteString, track: str) -> str:
    if (len(track) % 16) == 0:
        block = track + "8000000000000000"
    else:
        block = track + "80" + "0" * (16 - len(track) % 16 - 2)
    ic(block)

    mac = des.mac_2_ede(udk, ByteString(block))
    ic(mac)

    return str(mac[-2:])


def generate_cvc3(udk: ByteString, block: str) -> str:
    cryptogram = des.tdea_2_ede(udk, ByteString(block))
    ic(cryptogram)
    return F"{cryptogram[-2:].int:05d}"
